
/* _utf_8b.c: python utf-8b extension module */

/*

Copyright (c) 2006, Eric Tiedemann
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:


    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    * Neither the name of Eric Tiedemann nor the names of his
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "Python.h"

#include "utf_8b.h"

typedef struct {
	PyObject_HEAD
	struct utf_8b_decoder decoder;
} UTF_8bDecoderObject;

static PyTypeObject UTF_8bDecoder_Type;

#define UTF_8bDecoderObject_Check(v)	((v)->ob_type == &UTF_8bDecoder_Type)

static UTF_8bDecoderObject *
newUTF_8bDecoderObject(PyObject *arg)
{
	UTF_8bDecoderObject *self;
        self = (UTF_8bDecoderObject *) PyObject_New(UTF_8bDecoderObject, 
						    &UTF_8bDecoder_Type);
	if (self == NULL) {
	  return NULL;
	}

	utf_8b_decoder_init(&(self->decoder));

	return self;
}

/* UTF_8bDecoder methods */
static void
UTF_8bDecoder_dealloc(UTF_8bDecoderObject *self)
{
	PyObject_Del(self);
}

static PyObject *
UTF_8bDecoder_call(UTF_8bDecoderObject *self, PyObject *args, PyObject *kwds)
{
  char *s;
  int n;
  utf_8b_codepoint *out;
  int nout = 0;
  PyObject *pyout;
  int i;

  /* FIXME check that we have no keywords */
  if (!PyArg_ParseTuple(args, "s#:_utf_8b.UTF_8bDecoder.__call__", &s, &n))
    return NULL;

  out = malloc(n * sizeof(utf_8b_codepoint));

  if (out == NULL) {
    PyErr_NoMemory();
    return NULL;
  }
  
  for (i = 0; i < n; i++) {
    int rc = utf_8b_decoder_put(&(self->decoder),
				(utf_8b_byte) s[i], 
				out + nout);
    nout += rc;
  }
  
  /* check for codepoints too wide for a narrow python build */
#if Py_UNICODE_SIZE < 4
  for (i = 0; i < nout; i++) {
    utf_8b_codepoint c = out[i];
    if (c > 0xFFFF) {
      char buf[200];
      snprintf(buf, sizeof(buf),
	       "decoded codepoint U%8.8X not in range(0x10000) "
	       "(narrow Python build)", c);
      free(out);
      PyErr_SetString(PyExc_ValueError, buf);
      return NULL;
    }
  }
#endif

  pyout = PyUnicode_FromUnicode(NULL, nout);

  if (pyout == NULL) {
    free(out);
    PyErr_NoMemory();
    return NULL;
  }

  for (i = 0; i < nout; i++) {
    PyUnicode_AS_UNICODE(pyout)[i] = out[i];
  }

  free(out);

  return Py_BuildValue("Ni", pyout, n);
}


static PyTypeObject UTF_8bDecoder_Type = {
	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	"_utf_8b.UTF_8bDecoder",		/*tp_name*/
	sizeof(UTF_8bDecoderObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)UTF_8bDecoder_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
        0,                      /*tp_getattr*/
	0,                      /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
        (ternaryfunc)UTF_8bDecoder_call,     /*tp_call*/
        0,                      /*tp_str*/
        0,                      /*tp_getattro*/
        0,                      /*tp_setattro*/
        0,                      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,     /*tp_flags*/
        0,                      /*tp_doc*/
        0,                      /*tp_traverse*/
        0,                      /*tp_clear*/
        0,                      /*tp_richcompare*/
        0,                      /*tp_weaklistoffset*/
        0,                      /*tp_iter*/
        0,                      /*tp_iternext*/
        0,                      /*tp_methods*/
        0,                      /*tp_members*/
        0,                      /*tp_getset*/
        0,                      /*tp_base*/
        0,                      /*tp_dict*/
        0,                      /*tp_descr_get*/
        0,                      /*tp_descr_set*/
        0,                      /*tp_dictoffset*/
        0,                      /*tp_init*/
        0,                      /*tp_alloc*/
        0,                      /*tp_new*/
        0,                      /*tp_free*/
        0,                      /*tp_is_gc*/
};
/* --------------------------------------------------------------------- */

static PyObject *
_utf_8b_encode(PyObject *self, PyObject *args)
{
	Py_UNICODE *u;
	int n;
	unsigned char *b, *bp;
	int i;

	if (!PyArg_ParseTuple(args, "u#:_utf_8b.encode", &u, &n))
		return NULL;

	/* each input codepoint can encode to 0 to 4 output bytes */
	b = malloc(4 * n);
	if (b == NULL) {
	  PyErr_NoMemory();
	  return NULL;
	}

	bp = b;
	for (i = 0; i < n; i++) {
	  Py_UNICODE uc = u[i];
	  int rc = utf_8b_encode(uc, bp);
	  if (rc == 0) {
	    return NULL;
	  }
	  bp += rc;
	}

	return Py_BuildValue("s#i", b, bp - b, n);
}

/* Function of no arguments returning new UTF_8bDecoder object */
static PyObject *
_utf_8b_new(PyObject *self, PyObject *args)
{
	UTF_8bDecoderObject *rv;

	if (!PyArg_ParseTuple(args, ":decoder"))
		return NULL;
	rv = newUTF_8bDecoderObject(args);

	if (rv == NULL)
		return NULL;
	return (PyObject *)rv;
}

static PyMethodDef _utf_8b_methods[] = {
	{"encode",		_utf_8b_encode,		METH_VARARGS,
	 	NULL},
	{"decoder",		_utf_8b_new,		METH_VARARGS,
		PyDoc_STR("decoder() -> new UTF_8bDecoder object")},
	{NULL,		NULL}		/* sentinel */
};


PyDoc_STRVAR(module_doc,
 "UTF-_8b codec\n"
 "\n"
 " encode(<unicode string>) -> (<byte string>, <unichrs consumed>)\n"
 " decoder() -> <decoder object>\n"
 " <decoder object>(<byte string>) -> (<unicode string>, <bytes consumed>)\n");

PyMODINIT_FUNC
init_utf_8b(void)
{
	if (PyType_Ready(&UTF_8bDecoder_Type) < 0)
		return;

	Py_InitModule3("_utf_8b", _utf_8b_methods, module_doc);
}
