/* utf_8b.c: UTF-8b codec */

/*

Copyright (c) 2006, Eric Tiedemann
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:


    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    * Neither the name of Eric Tiedemann nor the names of his
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "utf_8b.h"

    /*

      [ the words of Master Kuhn ]

      Emit a malformed UTF-16 sequence for every byte in a malformed 
      UTF-8 sequence.

      Malformed UTF-8 sequences consist excludively of the bytes 
      0x80 - 0xff, and each of these bytes can be represented using a
      16-bit value from the UTF-16 low-half surrogate zone U+DC80 to
      U+DCFF.

    */

int utf_8b_encode(utf_8b_codepoint c, utf_8b_byte *p) {
  if (c < 0x80) {
    *p = c;
    return 1;
  } else if (c < 0x800) {
    /* 110zzzzx 10xxxxxx */
    *p++ = 0xC0 | (c >> 6);
    *p++ = 0x80 | (c & 0x3F);
    return 2;
  } else if (c >= 0xDC00 && c <= 0xDCFF) {
    /* reverse utf-_8b encoding of garbage bytes */
    *p++ = c - 0xDC00;
    return 1;
  } else if (c >= 0xD800 && c <= 0xDFFF) {
    // other surrogates
    return 0;
  } else if (0) {
    // anything else we should indicate error on?
    return 0;
  } else if (c < 0x10000) {
    /* 1110zzzz 10zxxxxx 10xxxxxx */
    *p++ = 0xE0 | (c >> 12);
    *p++ = 0x80 | ((c >> 6) & 0x3F);
    *p++ = 0x80 | (c & 0x3F);
    return 3;
  } else if (c <= UTF_8B_MAX_CODEPOINT) {
    /* 11110zzz 10zzxxxx 10xxxxxx 10xxxxxx */
    *p++ = 0xF0 | (c >> 18);
    *p++ = 0x80 | ((c >> 12) & 0x3F);
    *p++ = 0x80 | ((c >> 6) & 0x3F);
    *p++ = 0x80 | (c & 0x3F);
    return 4;
  } else {
    return 0;
  }
}

void utf_8b_decoder_init(struct utf_8b_decoder *p) {
  p->c = 0;
  p->expecting = 0;
  p->recordi = 0;
  p->zmask = 0;
}

void _utf_8b_bad(struct utf_8b_decoder *p) {
  fprintf(stderr, "utf-8 decoding trouble on the following byte sequence:\n");
  int i = 0;
  for (; i < p->recordi; i++) {
    fprintf(stderr, "%2.2x\n", p->record[i]);
  }
}

int utf_8b_decoder_invalid_bytes(struct utf_8b_decoder *p, 
				utf_8b_codepoint *out) {
  int n = p->recordi;
  p->expecting = p->recordi = 0;
  int i = 0;
  for (; i < n; i++) {
    *out++ = 0xDC00 + p->record[i];
  }
  return n;
}

int utf_8b_decoder_put(struct utf_8b_decoder *p, 
		      utf_8b_byte b, utf_8b_codepoint *out) {
  /*fprintf(stderr, "%x %x %d %d\n", b, c, expecting, recordi); */
  if (!p->expecting) {
    /* first byte */
    if (b < 0x80) {
      p->c = b;
      *out = p->c;
      return 1;
    } else {
      p->record[0] = b;
      p->recordi = 1;
      
      if ((b & 0xE0) == 0xC0) {
	/* 110zzzzx 10xxxxxx */
	if ((b & 0x1E) == 0)
	  return utf_8b_decoder_invalid_bytes(p, out);
	p->expecting = 1;
	p->zmask = 0;
	p->c = b & 0x1F;
      } else if ((b & 0xF0) == 0xE0) {
	/* 1110zzzz 10zxxxxx 10xxxxxx */
	p->expecting = 2;
	if ((b & 0xF) == 0)
	  p->zmask = 0x20;
	else
	  p->zmask = 0;
	p->c = b & 0xF;
      } else if ((b & 0xF8) == 0xF0) {
	/* 11110zzz 10zzxxxx 10xxxxxx 10xxxxxx */
	p->expecting = 3;
	if ((b & 0x7) == 0)
	  p->zmask = 0x30;
	else
	  p->zmask = 0;
	p->c = b & 7;
      } else {
	return utf_8b_decoder_invalid_bytes(p, out);
      }
    }
  } else {
    /* expecting */
    p->record[p->recordi++] = b;
    if (p->zmask && (b & p->zmask) == 0) {
      return utf_8b_decoder_invalid_bytes(p, out);
    }
    /* zmask is used for the 2nd byte only */
    p->zmask = 0;
    /* all the continuation bytes should be of the form 10xxxxxx */
    if ((b & 0xC0) != 0x80) {
      return utf_8b_decoder_invalid_bytes(p, out);
    }
    p->c = (p->c << 6) | (b & 0x3F);
    p->expecting--;
    if (p->expecting == 0) {
      /* aha!  we must still check that we haven't concocted an invalid */
      /* code-point..for example, one greater than MAX_CODEPOINT. */
      if (p->c > UTF_8B_MAX_CODEPOINT) {
	return utf_8b_decoder_invalid_bytes(p, out);
      } else if (p->c >= 0xD800 && p->c <= 0xDFFF) {
	/* ..or naughty surrogate pairs */
	return utf_8b_decoder_invalid_bytes(p, out);
      } else {
	p->expecting = p->recordi = 0;
	*out = p->c;
	return 1;
      }
    }
  }
  
  return 0;
}
